﻿

'
'
'/**
'--------------------------------------------------
' Version .....: 18/02/2010 15:40:41   
'               [dd/mm/aaaa] 
' Original en .: SerializaTreeView
'--------------------------------------------------
'
' Este archivo contiene dos objetos
'
'  Clase [LoadSaveArbol] 
'        que contiene la logica para guardar / leer un arbol en disco
'
'  Estructura [EstructruraNodoSerialiar] 
'        Que se emplea para volcar los datos 
'        (que nos nteresa guardar del nodo del arbol) a una linea
'        de  texto con formato CSV y al reves, recuperar los
'        datos de  una linea CSV
'--------------------------------------------------
'*/

'
Option Explicit On
Option Strict On
Option Infer Off
Option Compare Binary

Imports System.Windows.Forms

#Region "Clase LoadSaveArbol"



'
'
'/**
'--------------------------------------------------
' Version .....: 18/02/2010 15:40:41   
'               [dd/mm/aaaa] 
' Original en .: SerializaTreeView
'--------------------------------------------------
'
'
' -------------------------------------------------
' Clase ....: LoadSaveArbol
' Lenguaje .: (.NET Framework 3.5) - Visual Basic .NET (Version 9.0) - 2010
' Autor ....:  - Aguila - Joaquin
' Emilio ...: mailto:joaquin@medina.name
'
' DESCRIPCION.:
'    serializa, deserializa un arbol en unfichero
'
' OBSERVACIONES.:
'     Sin la referecia [Microsoft.VisualBasic] Sin Usar ControlChars
'
' BIBLIOGRAFIA.:
'    ### Falta la Bibliografía ###
'
' HERENCIA.:
'    ### Falta la Cadena completa de herencias ###
'
' MODIFICACIONES [dd/mm/aaaa] 
' - 18/02/2010 15:40:41 - Creación 
' -------------------------------------------------
'
'*/

''' <summary>
'''  Se encarga de escribir o de leer un arbol en un fichero
''' </summary>
''' <remarks></remarks>
Public Class LoadSaveArbol




    Dim codificacion As System.Text.Encoding = System.Text.Encoding.UTF32
    '   Dim nombreFichero As String = "D:/borrar/arbol.txt"


#Region "Save"

    ''' <summary>
    ''' Guarda un arbol en un fichero en formato CSV
    ''' </summary>
    ''' <param name="unArbol">el arbol que hay que guardar</param>
    ''' <param name="nombreDeFichero">el nombre completo del fichero</param>
    ''' <remarks></remarks>
    Public Sub Save( _
           ByVal unArbol As System.Windows.Forms.TreeView, _
           ByVal nombreDeFichero As String)

        '-------------------------------------
        ' para evitar errores tontos
        ' si el directorio no existe crearlo
        If IO.Directory.Exists(IO.Path.GetDirectoryName(nombreDeFichero)) = False Then
            IO.Directory.CreateDirectory(IO.Path.GetDirectoryName(nombreDeFichero))
        End If
        '-------------------------------------
        ' Recorrer los nodos raices del arbol
        Using SW As New System.IO.StreamWriter(nombreDeFichero, False, codificacion)
            For Each TreeNode As TreeNode In unArbol.Nodes
                AuxSaveRecorrerNodosHijos(0, TreeNode, SW)
            Next
        End Using

    End Sub


    ''' <summary>
    '''  Funcion recursiva que recorre los nodos hijos de cada nodo
    ''' </summary>
    ''' <param name="IDNodoPadre">
    ''' El ID del nodo padre</param>
    ''' <param name="nodo">
    ''' El Objeto [TreeNode] que se esta estudiando </param>
    ''' <param name="SW">
    ''' El objeto [StreamWriter] usado para grabar la informacion en el disco
    ''' </param>
    ''' <remarks></remarks>
    Private Sub AuxSaveRecorrerNodosHijos( _
                ByVal IDNodoPadre As Integer, _
                ByVal nodo As TreeNode, _
                ByRef SW As System.IO.StreamWriter)
        '--------------------------------
        ' el contador para generar un numero correlativo de nodos
        Static contadorIDNodo As Integer = 0
        ' Calcular y asigbar el ID unico a este nodo
        contadorIDNodo += 1
        Dim IDEsteNodoHijo As Integer = contadorIDNodo
        ' almacenar los valores del nodo
        Dim estruNodo As New EstructruraNodoSerialiar(IDEsteNodoHijo, IDNodoPadre, nodo.Text)

        ' montar la cadena de salida con el salto de linea al final
#If DEBUG Then
        ' para control de depuracion
        Debug.Write(estruNodo.CSV)
#End If
        ' escribir el nodo en el fichero
        SW.Write(estruNodo.CSV)

        ' recorrer los nodos hijos
        For Each unNodo As TreeNode In nodo.Nodes
            AuxSaveRecorrerNodosHijos(IDEsteNodoHijo, unNodo, SW)
        Next
    End Sub

#End Region


#Region "Load"

    ''' <summary>
    '''  Deserializa un árbol guardado en un fichero de texto
    ''' </summary>
    ''' <param name="outputArbolSalida">
    ''' El árbol que se va a cargar con la información 
    ''' que existe en el fichero de texto</param>
    ''' <param name="nombreDeFichero">
    ''' El nombre del fichero en el disco que contiene 
    ''' la información del árbol</param>
    Public Sub Load( _
               ByRef outputArbolSalida As System.Windows.Forms.TreeView, _
               ByVal nombreDeFichero As String)
        Try
            '----------------------------------------
            ' PASO UNO cargar el texto en una lista
            '----------------------------------------
            Dim listaNodos As New List(Of EstructruraNodoSerialiar)
            listaNodos = AuxLoadCargarLista(nombreDeFichero)

            '----------------------------------------
            ' PASO DOS cargar el arbol a partir de la lista
            ' busco el nodo raiz (o los nodos) lo grabo 
            ' y despues busco sus hijos
            '---------------------------------------
            Call AuxLoadCargaNodosRaiz(outputArbolSalida, listaNodos)
        Catch ex As Exception
            Throw
        End Try
    End Sub




    ''' <summary>
    '''    Lee en la lista de nodos los nodos raíces, 
    '''    los cuelga del árbol y a continuación busca 
    '''    los nodos hijos de ese nodo raíz
    ''' </summary>
    ''' <param name="outputArbolSalida">
    '''    El árbol que se va a cargar con la información 
    '''    que existe en la lista genérica</param>
    ''' <param name="listaNodos">
    '''    La lista genérica que contiene las estructuras 
    '''    con la información de cada nodo</param>
    Private Sub AuxLoadCargaNodosRaiz( _
                ByRef outputArbolSalida As System.Windows.Forms.TreeView, _
                ByRef listaNodos As List(Of EstructruraNodoSerialiar))


        Dim estrucNodoPadre As EstructruraNodoSerialiar = Nothing
        '----------------------------------------
        'busco el nodo raíz (o los nodos) lo grabo y después busco sus hijos
        '----------------------------------------
        outputArbolSalida.Nodes.Clear()
        Dim nuevoNodoRaiz As TreeNode
        For indexLista As Integer = 0 To listaNodos.Count - 1
            estrucNodoPadre = listaNodos.Item(indexLista)
            ' si no esta grabado
            If estrucNodoPadre.Grabado = False Then
                ' si es un nodo raiz
                If estrucNodoPadre.IDPadre = 0 Then
                    nuevoNodoRaiz = New TreeNode(estrucNodoPadre.Text)
                    outputArbolSalida.Nodes.Add(nuevoNodoRaiz)
#If DEBUG Then
                    ' control de depuracion
                    Debug.Write("Alta Raiz " & estrucNodoPadre.ToString)
#End If

                    ' mmarcarlo como grabado
                    estrucNodoPadre.Grabado = True
                    ' buscar y grabar sus hijos
                    AuxLoadCargaNodosHijos(nuevoNodoRaiz, estrucNodoPadre, listaNodos)
                End If

            End If

        Next

    End Sub



    ''' <summary>
    '''    Busca los nodos hijos de un nodo en la lista de nodos, 
    '''    cuando encuentra uno, lo cuelga del árbol, 
    '''    lo marca como colgado del árbol, 
    '''    y de forma recursiva busca sus hijos
    ''' </summary>
    ''' <param name="nodoPadre">
    '''    El objeto nodo padre, es decir el nodo del cual 
    '''    estamos buscando sus hijos</param>
    ''' <param name="estrucNodoPadre">
    '''    La estructura que contiene la información del nodo padre. 
    '''    Se pasa la estructuras porque necesito conocer 
    '''    los ID de los nodos</param>
    ''' <param name="listaNodos">
    '''    La lista genérica que contiene las estructuras 
    '''    con la información de cada nodo</param>
    Private Sub AuxLoadCargaNodosHijos( _
                    ByRef nodoPadre As TreeNode, _
                    ByRef estrucNodoPadre As EstructruraNodoSerialiar, _
                    ByRef listaNodos As List(Of EstructruraNodoSerialiar))

        ' a partir de ese nodo buscar los hijos
        'Dim NodoSalida As TreeNode = Nothing
        For indexLista As Integer = estrucNodoPadre.IDPadre + 1 To listaNodos.Count - 1
            Dim estrucNodoHijo As EstructruraNodoSerialiar = Nothing
            estrucNodoHijo = listaNodos.Item(indexLista)
            ' si no esta grabado Grabarlo como hijo de su padre
            If estrucNodoHijo.Grabado = False Then
                If estrucNodoHijo.IDPadre = estrucNodoPadre.ID Then

                    Dim nuevoNodoHijo As New TreeNode(estrucNodoHijo.Text)
                    nodoPadre.Nodes.Add(nuevoNodoHijo)
                    ' control de depuracion
#If DEBUG Then
                    Debug.Write("Alta Nodo " & estrucNodoHijo.ToString)
#End If

                    ' mmarcarlo como grabado
                    estrucNodoHijo.Grabado = True

                    ' buscar sus hijos
                    ' llamada recursiva a esta funcion
                    AuxLoadCargaNodosHijos(nuevoNodoHijo, estrucNodoHijo, listaNodos)
                End If
            End If
        Next

    End Sub



    ''' <summary>
    ''' Esta función lee el fichero de texto y carga 
    ''' la información en una lista genérica
    ''' </summary>
    ''' <param name="nombreDeFichero">
    ''' El nombre del fichero de texto donde está 
    ''' la información del árbol serializado
    ''' </param>
    ''' <returns>
    ''' Devuelve una lista genérica que contiene estructuras 
    ''' de datos del tipo [EstructruraNodoSerialiar] de manera 
    ''' que cada una de las estructuras contiene la información 
    ''' de un nodo del árbol. En otro proceso, usando esa 
    ''' información, se reconstruye el árbol
    ''' </returns>
    Private Function AuxLoadCargarLista(ByVal nombreDeFichero As String) As List(Of EstructruraNodoSerialiar)

        '-----------------------------------------------
        ' control de la existencia del fichero en el disco
        If IO.File.Exists(nombreDeFichero) = False Then
            Throw New System.IO.FileNotFoundException( _
            "El fichero no existe en el disco", nombreDeFichero)
        End If

        '----------------------------------------
        Dim listaNodos As New List(Of EstructruraNodoSerialiar)
        Dim estructuraUnNodo As EstructruraNodoSerialiar = Nothing



#If DEBUG Then
        Debug.WriteLine("*********************************")
        Debug.WriteLine("     | ID  | ")
        Debug.WriteLine(" ID  |Padre| Text")
        Debug.WriteLine("---------------------------------")
#End If

        '----------------------------------------
        ' cargar la lista leyendo el fichero de texto
        Using SR As New System.IO.StreamReader(nombreDeFichero, codificacion, True)
            ' leer la linea del fichero
            Dim input As String
            input = SR.ReadLine()
            While Not input Is Nothing

#If DEBUG Then
                Debug.WriteLine(input)
#End If
                ' escribir el nodo en un fichero
                estructuraUnNodo.CSV = input
                listaNodos.Add(estructuraUnNodo)
                input = SR.ReadLine()
            End While
            SR.Close()
        End Using
        Return listaNodos
    End Function

#End Region


#Region "Carga DatosEjemplo"

    Friend Shared Sub CargaArbolDatosEjemplo( _
           ByRef arbolSalida As System.Windows.Forms.TreeView)

        arbolSalida.Nodes.Clear()
        Dim raiz As TreeNode = arbolSalida.Nodes.Add("1.- Animales")

        Dim nodoMamiferos As TreeNode = raiz.Nodes.Add("1.1- Mamíferos")
        Dim nodoPerros As TreeNode = nodoMamiferos.Nodes.Add("1.1.1- Perros")
        nodoPerros.Nodes.Add("1.1.1.1- Pastores")
        nodoPerros.Nodes.Add("1.1.1.2- Chiguaguas")
        nodoPerros.Nodes.Add("1.1.1.3- Compañía")

        raiz.Nodes.Add("1.2-Serpientes")

        Dim nodoAves As TreeNode = raiz.Nodes.Add("1.3-Aves")
        nodoAves.Nodes.Add("1.3.1- Voladoras")
        nodoAves.Nodes.Add("1.3.2- Caminantes")

        Dim nodoPeces As TreeNode = raiz.Nodes.Add("1.4-Peces")
        Dim nodoPecesLuchadores As TreeNode = nodoPeces.Nodes.Add("1.4.1- Luchadores")
        nodoPecesLuchadores.Nodes.Add("1.4.1.1- Azules")
        nodoPecesLuchadores.Nodes.Add("1.4.1.2- Colorados")
        nodoPeces.Nodes.Add("1.4.2- Gupis")
        nodoPeces.Nodes.Add("1.4.3- Truchas")

        raiz = arbolSalida.Nodes.Add("2.- Hongos")

        raiz = arbolSalida.Nodes.Add("3.- Plantas")
        raiz.Nodes.Add("3.1- Hierbas")
        raiz.Nodes.Add("3.2- Arbustos")
        raiz.Nodes.Add("3.3- Arboles")


        raiz = arbolSalida.Nodes.Add("4.- Minerales")
        raiz.Nodes.Add("4.1- Piedras")


    End Sub


#End Region




End Class


#End Region

#Region "Estructura EstructruraNodoSerialiar"

'
' 
'
'/**
'--------------------------------------------------
' Version .....: 18/02/2010 15:40:41   
'               [dd/mm/aaaa] 
' Original en .: SerializaTreeView
'--------------------------------------------------
'
'
' -------------------------------------------------
' Clase ....: EstructruraNodoSerialiar
' Lenguaje .: (.NET Framework 3.5) - Visual Basic .NET (Version 9.0) - 2010
' Autor ....:  - Aguila - Joaquin
' Emilio ...: mailto:joaquin@medina.name
'
' DESCRIPCION.:
'    Estructura empleada para serializar un nodo de un arbol [System.Windows.Forms.TreeView]
'
' OBSERVACIONES.:
'     Sin la referecia [Microsoft.VisualBasic] Sin Usar ControlChars
'
' La estructura contiene los valores que se necesitan para volver a generar
' el árbol en la situación inicial, el ID ÚNICO de cada nodo, el ID del nodo 
' padre, el valor del campo [Text] (se puede guardar mas información si 
' fuera necesario paro no influye en el proceso) y un valor lógico que 
' se usa únicamente en el momento de reconstrucción del árbol, e indica 
' si este valor ya se ha dado de alta en el árbol o no
'
' La idea es que al cargar la estructura con la información, la propia 
' estructura genera una cadena CSV con los datos que se guardan 
' en un fichero de texto
'
' En el momento de reconstruir el árbol, lo primero que se hace es leer 
' el texto, y por cada línea se carga una estructura, que luego se guarda 
' en una lista genérica. Por último se recorre la lista y se van creando 
' cada uno de los nodos en el árbol con la información que figura en 
' cada una de las estructuras (que contienen la información de un nodo) 
' que figuran en la  lista.
'
' La estructura contiene un campo especial [Grabado] que se usa en el 
' proceso de reconstrucción del árbol, y se emplea para indicar si ese 
' nodo se ha dado de alta o no. Es una necesidad que se deriva del  
' algoritmo elegido para reconstruir el árbol. El algoritmo recorre 
' varias veces secuencialmente la lista para localizar los hijos de 
' un nodo padre y conforme se encuentran se van dado de alta, cuando 
' se vuelva a leer ese nodo este campo indica que ya se dio de alta 
' a el y a todos sus hijos y que por lo tanto podemos saltárnoslo y 
' leer el siguiente elemento (estructura)  de la lista
'
' BIBLIOGRAFIA.:
'    ### Falta la Bibliografía ###
'
' HERENCIA.:
'    ### Falta la Cadena completa de herencias ###
'
' MODIFICACIONES [dd/mm/aaaa] 
' - 18/02/2010 15:40:41 - Creación 
' -------------------------------------------------
'
'*/
'

' -------------------------------------------------
''' <summary>
'''  Estructura empleada para serializar un nodo de un arbol [System.Windows.Forms.TreeView]
''' </summary>
Public Structure EstructruraNodoSerialiar

    ''' <summary> El id UNICO del nodo </summary>
    Private _ID As Integer
    ''' <summary> El id del nodo padre </summary>
    Private _IDPadre As Integer
    ''' <summary> El valor de la propiedad Text del nodo </summary>
    Private _Text As String
    ''' <summary>
    ''' Es un valor lógico que indica si el nodo se ha creado en el árbol (valor true) o no
    ''' </summary>
    Private _Grabado As Boolean

    ''' <summary>
    '''  El id UNICO del nodo
    ''' </summary>
    Public Property ID() As Integer
        Get
            Return _ID
        End Get
        Private Set(ByVal value As Integer)
            _ID = value
        End Set
    End Property

    ''' <summary>
    '''  El id del nodo padre
    ''' </summary>
    Public Property IDPadre() As Integer
        Get
            Return _IDPadre
        End Get
        Private Set(ByVal value As Integer)
            _IDPadre = value
        End Set
    End Property

    ''' <summary>
    '''  El valor de la propiedad Text del nodo
    ''' </summary>
    Public Property Text() As String
        Get
            Return _Text
        End Get
        Private Set(ByVal value As String)
            ' quitar los espacios en blanco de la cadena
            _Text = value.Trim
        End Set
    End Property

    ''' <summary>
    ''' Se emplea en el proceso de carga del árbol, 
    ''' y es un valor lógico que indica si el nodo se ha creado en el árbol (valor true) o no
    ''' </summary>
    ''' <returns>
    ''' Valor lógico que indica si el nodo se ha creado en el árbol o aun no
    ''' TRUE, el nodo ya se creó en el árbol
    ''' FALSE el nodo aun no se ha dado de alta en el árbol
    ''' </returns>
    ''' <remarks></remarks>
    Public Property Grabado() As Boolean
        Get
            Return _Grabado
        End Get
        Set(ByVal value As Boolean)
            _Grabado = value
        End Set
    End Property


    ''' <summary>
    '''  Constructor de la estructra
    ''' </summary>
    ''' <param name="id">El id UNICO del nodo</param>
    ''' <param name="idPadre"> El id del nodo padre</param>
    ''' <param name="text"> El valor de la propiedad Text del nodo</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal id As Integer, ByVal idPadre As Integer, ByVal text As String)
        Me.ID = id
        Me.IDPadre = idPadre
        Me.Text = text
        Me.Grabado = False
    End Sub


    ''' <summary>
    ''' Devuelve los datos de la estructura en un formato legible por el usuario
    ''' </summary>
    ''' <returns>Una cadena con la información de la estructura 
    ''' por ejemplo ID = 5; ID Padre = 2; Text = pajaros; Grabado = False
    ''' </returns>
    Public Overrides Function ToString() As String
        'La sintaxis de un elemento de formato es {index[,alignment][:formatString]},
        Dim provider As IFormatProvider
        provider = System.Globalization.CultureInfo.CurrentCulture
        Dim formato As String = " ID ={0,5:0}; ID Padre = {1,5:0}; Text = {2,-20}; Grabado = {3,-6}"
        Dim salida As String
        ' montar la cadena de salida con el salto de linea al final
        salida = String.Format(provider, formato, ID, IDPadre, Text, Grabado) & Environment.NewLine
        Return salida
    End Function


    ''' <summary>
    ''' Obtiene / establece los valores de la estructura 
    ''' en una cadena con formato CSV
    ''' </summary>
    Public Property CSV() As String
        Get
            Dim provider As IFormatProvider
            provider = System.Globalization.CultureInfo.CurrentCulture
            'La sintaxis de un elemento de formato es
            ' {index[,alignment][:formatString]},
            Dim formato As String = "{0,5:0}|{1,5:0}|{2,-20}"
            Dim salida As String
            '----------------------
            ' montar la cadena de salida con el salto de linea al final
            salida = String.Format(provider, formato, ID, IDPadre, Text) & Environment.NewLine
            '----------------------
            Return salida
        End Get
        Set(ByVal value As String)
            If String.IsNullOrEmpty(value) Then
                Throw New ArgumentException(" el valor de la cadena no puede ser nulo o vacio", "CSV")
            End If
            Dim matrizElementos() As String
            Dim matrizSeparadores() As Char = {"|"c}
            Try
                matrizElementos = value.Split(matrizSeparadores, StringSplitOptions.RemoveEmptyEntries)
                Me.ID = Integer.Parse(matrizElementos(0))
                Me.IDPadre = Integer.Parse(matrizElementos(1))
                Me.Text = matrizElementos(2)
                Me.Grabado = False
            Catch ex As Exception
                Throw
            End Try
        End Set
    End Property

End Structure

#End Region
